(ns damp.ekeko
   (:refer-clojure :exclude [==])
   (:use clojure.core.logic)
  ; (:use clojure.inspector) 
   (:use [damp.ekeko jdtnode logic gui mli reification basic])
   (:use [damp.ekeko.util maps mattias])

   (:import [org.eclipse.core.runtime.jobs Job]
            [org.eclipse.core.resources ResourcesPlugin]))

(set! *warn-on-reflection* true)

(defmacro ekeko [logicvars & goals]
  `(run* [resultvar#] 
         (fresh [~@logicvars]
                (equals resultvar# [~@logicvars])
                ~@goals)))

(defmacro ekeko* [vars & goals]
  `(let [querystr# (str '(ekeko* [~@vars] ~@goals))
         start# (System/nanoTime)
         resultsqc# (run* [resultvar#] 
                          (fresh [~@vars]
                                 (equals resultvar# [~@vars])
                                 ~@goals))
         elapsed#  (/ (double (- (System/nanoTime) start#)) 1000000.0)
         cnt# (count resultsqc#)]
     (eclipse-uithread-return (fn [] (open-barista-results-viewer* querystr# '(~@vars) resultsqc# elapsed# cnt#)))))
             

;add build listener to wait for model to be populated before querying
;add it to getInstance on EkekoModel! 
;better: ensurePopulated method, called by ekeko macro


(defn workspace-wait-for-builds-to-finish []
  (.join (Job/getJobManager) (ResourcesPlugin/FAMILY_AUTO_BUILD) nil))

; (ekeko-workspace-sequentially!  (fn [p] (.getName p)))
(defn workspace-enable-ekeko-sequentially-and-do!
  [f]
  (map-workspace 
    (fn [p] 
      (workspace-project-open! p)
      (workspace-project-enable-ekeko! p)
      (workspace-wait-for-builds-to-finish)
      (let [result (f p)]
        (workspace-project-disable-ekeko! p)
        (workspace-project-close! p)
        result))))
     



;Example repl session
(comment
  (use 'damp.ekeko)
  (in-ns 'damp.ekeko)
 (ekeko* [?t ?b] (ast :TypeDeclaration ?t) (equals ?b (.resolveBinding ?t)))
  (ekeko* [?m] (ast :MethodDeclaration ?m))
 
  (ekeko* [?cu] (ast :CompilationUnit ?cu))
  (ekeko* [?cu] (ast :StringLiteral ?cu))
  (ekeko* [?cu ?key ?child] (ast :CompilationUnit ?cu) (has ?key ?cu ?child))
  
  (ekeko* [?cu ?key ?child ?shouldbecu ?shouldbekey] (ast :CompilationUnit ?cu) (has ?key ?cu ?child) (has ?shouldbekey ?shouldbecu ?child))
  (ekeko* [?node ?nodetype ?childtype ?child] (ast ?nodetype ?node) (has ?childtype ?node ?child))
  
  (ekeko* [?node ?child ?keyword] (child ?keyword ?node ?child)) 
  
  (ekeko* [?cu ?exp] (ast :MethodDeclaration ?cu) (child+ ?cu ?exp) (succeeds (actual-expression? ?exp)))   
  
  
  (ekeko* [?inv] (ast :MethodInvocation ?inv) (has :expression ?inv nil))
  
  (ekeko* [?inv ?child] (ast :MethodInvocation ?inv) (child :arguments ?inv ?child))

  (ekeko* [?m ?o] (ast :MethodDeclaration ?m) (fresh [?os]
                                                  (equals ?os (method-overriders ?m))
                                                  (contains ?os ?o)))
  
  (ekeko* [?i ?m] (ast :MethodInvocation ?i) (fresh [?ts]
                                                  (equals ?ts (invocation-targets ?i))
                                                  (contains ?ts ?m)))
  
  (ekeko* [?import] (ast :ImportDeclaration ?import))
  
  (ekeko* [?s ?name ?initExpression ?condition ?update ?mslist] (for-mattias ?s ?name ?initExpression ?condition ?update ?mslist))    
 
  (ekeko* [?import ?binding] (import-declaration-imports-binding ?import ?binding))
  
  (ekeko* [?n ?key ?binding] (ast-declares-binding  ?n ?key ?binding))
  
  (ekeko* [?import ?package] (import-declaration-imports-package ?import ?package))
  (ekeko* [?import ?package ?name] (import-declaration-imports-package ?import ?package) (equals ?name (.getName ?package)))
  (ekeko* [?import ?package ?components] (import-declaration-imports-package ?import ?package) (equals ?components (vec (.getNameComponents ?package))))
  
  (def 
    packages-sorted
    (let [imported-packages-per-project
        (workspace-enable-ekeko-sequentially-and-do!
          (fn [p] (set (map first 
                            (ekeko [?name]
                                   (fresh [?import ?package]
                                          (import-declaration-imports-package ?import ?package) 
                                          (equals ?name (.getName ?package))))))))
        projects-per-imported-package 
        (damp.ekeko.util.maps/transpose-multimap imported-packages-per-project)]
    (sort-by (fn [[k v]] (count v)) > projects-per-imported-package)))

  )


