(ns damp.ekeko.basic
  (:refer-clojure :exclude [==])
  (:require (clojure set))
  (:use clojure.core.logic)
  (:use [damp.ekeko logic mli jdtnode reification])
  (:import [org.eclipse.jdt.core.dom IPackageBinding ITypeBinding IVariableBinding IMethodBinding IAnnotationBinding IMemberValuePairBinding]))
  
   
;; Basic reasoning predicates
;; --------------------------


;; AST

(defn ast-resolveable [?key ?ast]
  (let  [keys (ekeko-keywords-for-resolveable-ast-classes)]
    (all
         (contains keys ?key)
         (ast ?key ?ast))))

(defn ast-declaration [?key ?ast]
  (let [declaration-keywords (ekeko-keywords-for-declaration-ast-classes)]
    (all 
      (contains declaration-keywords ?key)
      (ast ?key ?ast))))

(defn ast-resolveable-declaration [?key ?ast]
  (all 
    (ast-declaration ?key ?ast)
    (ast-resolveable ?key ?ast)))

;; Bindings 

;Quite expensive
(defn ast-declares-binding [?n ?key ?binding]
  (all 
    (!= ?binding nil)
    (!= ?n nil)
    (conda [(v+ ?binding) (all
                            (equals ?n (binding-to-declaration ?binding))
                            (ast ?key ?n))]
           [(v- ?binding) (all
                            (ast-resolveable-declaration ?key ?n)
                            (equals ?binding (.resolveBinding ?n)))])))

; is-*? predicates are non-relational!

(defn is-binding-package? [?binding]
  (succeeds (binding-package? ?binding)))

(defn is-binding-type? [?binding]
  (succeeds (binding-type? ?binding)))

(defn is-binding-variable? [?binding]
  (succeeds (binding-variable? ?binding)))

(defn is-binding-method? [?binding]
  (succeeds (binding-method? ?binding)))

(defn is-binding-annotation? [?binding]
  (succeeds (binding-annotation? ?binding)))

(defn is-binding-member-value-pair? [?binding]
  (succeeds (binding-member-value-pair? ?binding)))

 
;; Imports

(defn import-declaration-imports-binding [?import ?binding]
  (all
    (ast :ImportDeclaration ?import)
    (equals ?binding (.resolveBinding ?import))))


(defn import-declaration-imports-package [?import ?packagebinding]
  (fresh [?binding]
    (import-declaration-imports-binding ?import ?binding)
    (conda [(is-binding-package? ?binding) (== ?binding ?packagebinding)]
           [(is-binding-type? ?binding) (equals ?packagebinding (.getPackage ?binding))] 
           ;non-dynamic, static import of field and method
           [(is-binding-method? ?binding) (equals ?packagebinding (-> ?binding (.getDeclaringClass) (.getPackage)))]        
           [(is-binding-variable? ?binding) (equals ?packagebinding (-> ?binding (.getDeclaringClass) (.getPackage)))])))

           


(defn import-declaration-imports-binding-type [?import ?binding]
  (all
    (import-declaration-imports-binding ?import ?binding)
    (is-binding-type? ?binding)))








    

;; Types

;; Call Graph

;; Control Flow Graph