(ns damp.ekeko.mli
  (:import [org.eclipse.core.resources ResourcesPlugin IWorkspace]
           [org.eclipse.jdt.core  IMember IJavaElement ITypeHierarchy JavaCore IType IJavaModel IJavaProject IPackageFragment ICompilationUnit]
           [org.eclipse.jdt.core.dom ConstructorInvocation SuperConstructorInvocation ClassInstanceCreation SuperMethodInvocation IMethodBinding AnonymousClassDeclaration IBinding MethodDeclaration MethodInvocation ASTParser AST ASTRequestor CompilationUnit ASTNode TypeDeclaration]
           [org.eclipse.core.runtime Status]
           [org.eclipse.ui PlatformUI IWorkingSet IWorkingSetManager]
           [damp.ekeko TableGatheringVisitor]
           [damp.ekeko EkekoModel JavaProjectModel ProjectModel])
  (:use damp.ekeko.util.jobs)
  (:use [damp.ekeko jdtnode])
  (:use clojure.inspector))

; JDT 
; ---

(defn ^IWorkspace eclipse-workspace []
  (ResourcesPlugin/getWorkspace))

(defn ^IJavaModel jdt-model []
  (JavaCore/create (.getRoot  (eclipse-workspace))))

; Projects
; --------


(defn jdt-projects []
  (.getJavaProjects (jdt-model)))

(defn ^IJavaProject jdt-project-named [name]
  (.getJavaProject (jdt-model) name))

(defn jdt-projects-names []
  (map (fn [^IJavaProject p] (.getElementName p))
       (jdt-projects)))

;Use Ekeko JavaProjectModel instead
;(defn jdt-project-compilationunits [^IJavaProject project]
;  (apply concat
;         (for [^IPackageFragment fragment (.getPackageFragments project)]
;           (.getCompilationUnits fragment))))





; Working Sets
; ------------

(defn ^IWorkingSetManager jdt-workingset-manager []
  (.getWorkingSetManager (PlatformUI/getWorkbench)))

(defn jdt-workingset-named [name]
  (.getWorkingSet (jdt-workingset-manager) name))



; MLI
; ---

(defn ekeko-model []
  (EkekoModel/getInstance))

(defn project-models []
  (.getProjectModels (ekeko-model)))

(defn java-project-models []
  (filter (fn [project-model] (instance? JavaProjectModel project-model))
          (project-models)))

(defn icu-to-ast [icu]
  (.getCompilationUnit
    (.getJavaProjectModel (ekeko-model) (.getJavaProject icu))
    icu))


; Workspace Projects (including non-JDT)
; --------------------------------------


(defn workspace-projects []
  (-> 
    (eclipse-workspace)
    (.getRoot)
    (.getProjects)))


(defn workspace-project-open? [p]
  (.isOpen p))

(defn workspace-project-open! [p]
  (when-not 
    (workspace-project-open? p)
    (.open p nil)))
  
(defn workspace-project-close! [p]
  (when
    (workspace-project-open? p)
    (.close p nil)))  

(defn workspace-project-ekeko-enabled? [p]
  (.hasProjectModel (ekeko-model) p))

(defn workspace-project-toggle-ekeko! [p]
  (EkekoModel/toggleNature p))

(defn workspace-project-enable-ekeko! [p]
  (when-not 
    (workspace-project-ekeko-enabled? p)
    (workspace-project-toggle-ekeko! p)))
    
(defn workspace-project-disable-ekeko! [p]
  (when
    (workspace-project-ekeko-enabled? p)
    (workspace-project-toggle-ekeko! p)))

(defn for-each-workspace-project [f] 
  (doseq [p (workspace-projects)]
    (f p)))

(defn map-workspace [f]
  (reduce (fn [results p] (assoc results (.getName p) (f p))) {} (workspace-projects)))

(defn workspace-close-projects! []
  (for-each-workspace-project workspace-project-close!))



; MLI for Project
; ---------------

;; (defn jdt-project-parser [^IJavaProject project]
;;   "Returns a function that parses a sqc of ICUs and returns an icu->ast map."
;;   (let [parser (ASTParser/newParser AST/JLS3)]                
;;     (.setResolveBindings parser true)
;;     (.setKind parser (ASTParser/K_COMPILATION_UNIT))
;;     (.setProject parser project)
;;     (fn [icus]
;;       (let [icu-to-ast (atom {})
;;             requestor (proxy [ASTRequestor] []
;;                     (acceptAST [icu ast]                               
;;                     (swap! icu-to-ast assoc icu ast)))]
;;         (println "Parsing compilation units of project " (.getElementName project))
;;         (.createASTs parser (into-array ICompilationUnit icus) (make-array java.lang.String 0) requestor nil)
;;         @icu-to-ast))))



;; (defn ^TableGatheringVisitor tables-for-asts [asts]
;;   (let [visitor (TableGatheringVisitor.)]
;;     (println "Visiting asts ...")
;;     (doseq [^CompilationUnit ast asts]
;;       (.accept ast visitor))
;;     (println "... done.")
;;     visitor))


;; (defn make-mli-for-project [project]
;;   "Returns a map with all information about a project."
;;  (let [parser (jdt-project-parser project)
;;        icus (jdt-project-compilationunits project)
;;        icu-to-ast (parser icus)
;;        tables (tables-for-asts (vals icu-to-ast))   ;indexatie per cu houden, anders kan de participant niet meedoen, zo ook makkelijk om te paralleliseren per cu
;;        ]
;;    {:icu-to-ast icu-to-ast
;;     :compilationUnits (vals icu-to-ast)
;;     :typeDeclarations (seq (.typeDeclarations tables))
;;     :annotationTypeDeclarations (seq (.annotationTypeDeclarations tables))
;;     :enumDeclarations (seq (.enumDeclarations tables))
;;     :methodDeclarations (seq (.methodDeclarations tables))
;;     :anonymousClassDeclarations (seq (.anonymousClassDeclarations tables))
;;     :expressions (seq (.expressions tables))
;;     :statements (seq (.statements tables))
;;     }))

;; (defn pmli-compilationUnits [pmli]
;;   (:compilationUnits pmli))

;; (defn pmli-typeDeclarations [pmli]
;;   (:typeDeclarations pmli))

;; (defn pmli-annotationTypeDeclarations [pmli]
;;   (:annotationTypeDeclarations pmli))

;; (defn pmli-enumDeclarations [pmli]
;;   (:enumDeclarations pmli))

;; (defn pmli-methodDeclarations [pmli]
;;   (:methodDeclarations pmli))

;; (defn pmli-anonymousClassDeclarations [pmli]
;;   (:anonymousClassDeclarations pmli))

;; (defn ^CompilationUnit pmli-icu-to-ast [pmli icu]
;;   (get (:icu-to-ast pmli) icu))
  
;; ; MLI for WorkingSet
;; ; ------------------

;; (defn mli-for-workingset [^IWorkingSet workingset]
;;   ;map from project to mliforproject
;;   (let [projects (.getElements workingset)]
;;     (map make-mli-for-project projects)))

;; (def ^:dynamic *mlis-for-jdt* (atom nil))

;; (defn initialize-mli-for-workingset [workingset]
;;   (reset! *mlis-for-jdt* (mli-for-workingset workingset)))

;; (defn mlis-for-jdt []
;;   @*mlis-for-jdt*)

;; (defn initialize-mli-for-workingset-named [name]
;;   (initialize-mli-for-workingset (jdt-workingset-named name)))

;; (defn ^CompilationUnit icu-to-ast [icu]
;;   (some (fn [pmli] (pmli-icu-to-ast pmli icu)) (mlis-for-jdt)))





; JDT Type Hierachy
; -----------------

(defn type-declaration-to-itype [t]
  (-> t .resolveBinding .getJavaElement))

(defn itype-type-hierarchy [^IType it]
  ;(.newTypeHierarchy it nil))
  (.getTypeHierarchy (ekeko-model) it))

  
(defn ^ITypeHierarchy type-declaration-type-hierarchy [t]
  (-> t type-declaration-to-itype itype-type-hierarchy))

(defn type-declaration-subtypes-itypes [^TypeDeclaration t]
  (let [h (type-declaration-type-hierarchy t)]
    (.getSubtypes h (.getType h))))

(defn type-declaration-subclasses-itypes [^TypeDeclaration t]
  (let [h (type-declaration-type-hierarchy t)]
    (.getSubclasses h (.getType h))))

(defn ielement-to-declaration [^IMember ije] 
  (when-let [icu (.getCompilationUnit ije)] ;nil if itype came not from source
    (.findDeclaringNode
      ^CompilationUnit (icu-to-ast icu)
      ^String (.getKey ije))))


; JDT DOM Bindings
; ----------------

(defn binding-to-declaration [^IBinding b]
  (when-let [ije (.getJavaElement b)]
    (ielement-to-declaration ije)))
    
; JDT Call Graph
; --------------

(defn method-overriders [^MethodDeclaration m]
  (let [mbinding (.resolveBinding m)]
    (mapcat
      ;(apply concat (pmap
      (fn [itype] 
              (if-let [subclass (ielement-to-declaration itype)]
                (filter 
                  (fn [d] 
                    (and (instance? MethodDeclaration d)
                         (.overrides ^IMethodBinding (.resolveBinding ^MethodDeclaration d)  mbinding)))
                  (if
                    (instance? AnonymousClassDeclaration subclass)
                    (.bodyDeclarations ^AnonymousClassDeclaration subclass)
                    (.getMethods ^TypeDeclaration subclass)))
                []))
            (type-declaration-subclasses-itypes (.getParent m)))))
      
   
(defn targets-of-constructor-invocation [n]
  (if-let [cb (.resolveConstructorBinding n)]
   (if-let [m (binding-to-declaration cb)]
     [m]
     [])
   []))

(defn targets-of-invocation [i]
  (if-let [mb (.resolveMethodBinding i)]
    (if-let [static-target (binding-to-declaration mb)]
      (conj  (method-overriders static-target) static-target)
      [])
    []))

(defmulti invocation-targets class)
(defmethod invocation-targets MethodInvocation [i]
  (targets-of-invocation i))
(defmethod invocation-targets SuperMethodInvocation [i]
  (targets-of-invocation i))
(defmethod invocation-targets ClassInstanceCreation [n]
  (targets-of-constructor-invocation n))
(defmethod invocation-targets ConstructorInvocation [i]
  (targets-of-constructor-invocation i))
(defmethod invocation-targets SuperConstructorInvocation [i]
  (targets-of-constructor-invocation i))


;(defmulti with-invocations (fn [ast fun] (class ast)))
;(defmethod with-invocations MethodInvocation [ast fun]
;  (fun ast)
;  (doseq [child (node-children ast)]
;    (with-invocations child fun)))
;(defmethod with-invocations :default [ast fun]
;  (doseq [child (node-children ast)]
;    (with-invocations child fun)))

;todo: onderdeel van mli maken? ... beter werken via memoize library
;(defn invocation-targets-within-node [^ASTNode ast
;                                      invocation-to-methods 
;                                      method-to-invocations]
;  (let [filler (fn [retriever]
;                 (fn [invocation]
;                   (let [targets (retriever invocation)]
;                     (swap! invocation-to-methods assoc invocation targets)                   ;ipv assoc conj-assoc of zoiets, en een set gebruiken      
;                     (doseq [target targets]
;                       (swap! method-to-invocations assoc target invocation))
;                     true)))]
;    (do
;        (with-invocations ast (filler targets-of-invocation))
;        [invocation-to-methods 
;                                      method-to-invocations]
;        )))
