(ns damp.ekeko.util.search
  (:import [java.util PriorityQueue]))
  
;based on https://github.com/aria42/mochi/blob/master/src/mochi/search.clj

(defprotocol ISearchProblem
  (is-goal? [this state] "state goal test")
  (init-state [this] "initial state of problem")
  (successors [this state] "seq of [neighbor step-cost]"))

(defrecord SearchNode
  [path cost-so-far priority]
  java.lang.Object
  (toString [this] (format "%s %s" path cost-so-far priority)))

(defn- pq-new []
  (PriorityQueue. 20
                  (comparator
                    (fn [^SearchNode x ^SearchNode y]
                      (< (.priority x) (.priority y))))))

(defn- init-search-node [search-problem]
  (SearchNode. (list (init-state search-problem)) 0.0 0.0))

(defn- generic-search
  [search-problem priority-fn]
  (let [pq (doto (pq-new) (.add (init-search-node search-problem)))]	
    (loop [ignore nil]
      (let [node (.poll pq)
            path (.path node)
            state (first path)]
        (if
          (is-goal? search-problem state)
          (reverse path) 
          (recur
            (doseq [[succ cost] (successors search-problem state)
                    :let [g (+ (.cost-so-far node) cost)]]
              (.add pq (SearchNode. (cons succ path)
                                    g
                                    (priority-fn node succ cost))))))))))

(defn astar-search
  [search-problem heuristic-fn]
  (generic-search search-problem
                  (fn [node succ cost]
                    (+ (.cost-so-far node) cost (heuristic-fn succ)))))

(defn uniform-cost-search
  [search-problem]
  (astar-search search-problem (constantly 0.0)))

(comment
(defrecord GraphPathSearchProblem [graph start goal]
  ISearchProblem
  (is-goal? [this state] (= state goal))
  (init-state [this] start)
  (successors [this state] (get graph state [])))


  (uniform-cost-search
   (GraphPathSearchProblem.
    {:s [[:a 1.0],[:b 0.5]] :a [[:g 1.0]] :b [[:g 1.0]]} :s :g))
)