package damp.ekeko;

import java.util.Vector;

import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AnnotationTypeDeclaration;
import org.eclipse.jdt.core.dom.AnnotationTypeMemberDeclaration;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.EnumConstantDeclaration;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ImportDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.QualifiedName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.StructuralPropertyDescriptor;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;

public class TableGatheringVisitor extends ASTVisitor {
	
	public Vector<TypeDeclaration> typeDeclarations = new Vector<TypeDeclaration>();
	public Vector<EnumDeclaration> enumDeclarations = new Vector<EnumDeclaration>();
	public Vector<AnonymousClassDeclaration> anonymousClassDeclarations = new Vector<AnonymousClassDeclaration>();
	public Vector<MethodDeclaration> methodDeclarations = new Vector<MethodDeclaration>();
	public Vector<FieldDeclaration> fieldDeclarations = new Vector<FieldDeclaration>();
	public Vector<AnnotationTypeDeclaration> annotationTypeDeclarations = new Vector<AnnotationTypeDeclaration>();
	public Vector<Statement> statements = new Vector<Statement>();
	public Vector<Expression> expressions = new Vector<Expression>();
	public Vector<SingleVariableDeclaration> singleVariableDeclarations = new Vector<SingleVariableDeclaration>();
	public Vector<EnumConstantDeclaration> enumConstantDeclarations = new Vector<EnumConstantDeclaration>();
	public Vector<AnnotationTypeMemberDeclaration> annotationTypeMemberDeclarations = new Vector<AnnotationTypeMemberDeclaration>();

	public boolean visit(AnnotationTypeMemberDeclaration atmd) {
		annotationTypeMemberDeclarations.add(atmd);
		return true;
	}
			
	public boolean visit(TypeDeclaration td) {
		typeDeclarations.add(td);
		return true;
	}

	public boolean visit(FieldDeclaration fd) {
		fieldDeclarations.add(fd);
		return true;
	}
	
	public boolean visit(EnumConstantDeclaration ecd) {
		enumConstantDeclarations.add(ecd);
		return true;
	}

	
	public boolean visit(SingleVariableDeclaration svd) {
		singleVariableDeclarations.add(svd);
		return true;
	}
	
	public boolean visit(AnnotationTypeDeclaration atd) {
		annotationTypeDeclarations.add(atd);
		return true;
	}
	
	public boolean visit(MethodDeclaration md) {
		methodDeclarations.add(md);
		return true;
	}
	
	public boolean visit(AnonymousClassDeclaration ac) {
		anonymousClassDeclarations.add(ac);
		return true;
	}
	
	public boolean visit(EnumDeclaration ec) {
		enumDeclarations.add(ec);
		return true;
	}

	public boolean visit(ImportDeclaration id) { 
		//See comment in preVisit
		return false;
	}

	public void preVisit(ASTNode n) {

		if (n instanceof Expression) {
			if (n instanceof Name) {
				Name nam = (Name) n;
				//Try to weed out as many as possible without having to resolve the name
				//See remark above nam.resolveBinding
				StructuralPropertyDescriptor locationInParent = nam.getLocationInParent();
				if (locationInParent.equals(QualifiedName.QUALIFIER_PROPERTY))
					return;
				if (locationInParent.equals(MethodInvocation.NAME_PROPERTY))//costly to resolve
					return;
				if (locationInParent.equals(MethodDeclaration.NAME_PROPERTY))//costly to resolve
					return;
				//There are name expressions in JEdit for which the next line returns a NPE (seems to be within import declarations only)
				IBinding b = nam.resolveBinding();
				if(b != null)
					if(IBinding.VARIABLE != b.getKind())
						return;
			}
			expressions.add((Expression) n);
			return;
		}
		if (n instanceof Statement) {
			if (n instanceof Block) 
				if(n.getParent() instanceof MethodDeclaration)
					return;
			statements.add((Statement) n);
			return;
		}
	}

}
