(ns damp.ekeko.reification
  (:refer-clojure :exclude [==])
  (:use [damp.ekeko logic mli jdtnode])
  (:use [clojure.core.logic])
  (:import [org.eclipse.jdt.core.dom ASTNode ASTNode$NodeList CompilationUnit]
           ))

;Developed against clojure 1.2, core.logic 0.6.5
;todo: newest core.logic no longer uses ?var names, [] around conde clauses

(declare nodes-of-type)

;; (ast ?classkeyword ?node) relation
;; ----------------------------------

;this might be slow due to reflection, but avoids having to import the entire package in order to refer to the classes as literals
;maybe it is faster to check in ast/2 whether the class name of a node is the same as the given classymbol


;todo: macro (cond+-e [?k ?n] ((:+ :-) ..

(defn ast [?keyword ?node]
  (conda [(v+ ?node) (conda [(v+ ?keyword) (succeeds-without-exception (instance? (class-for-ekeko-keyword ?keyword) ?node))]
                            [(v- ?keyword) (all (succeeds (instance? ASTNode ?node))
                                                (equals ?keyword (ekeko-keyword-for-class-of ?node)))])]
         [(v- ?node) (conda [(v+ ?keyword) (fresh [?nodes]
                                                  (equals ?nodes (nodes-of-type ?keyword))
                                                  (contains ?nodes ?node))]
                            [(v- ?keyword) (fresh [?keywords]
                                                  (equals ?keywords (map (fn [^Class c]
                                                                       (ekeko-keyword-for-class c))
                                                                     (node-classes)))
                                                  (contains ?keywords ?keyword)
                                                  (ast ?keyword ?node))])]))

;; (has ?keyword ?node ?child) relation
;; ----------------------------------------

(defn has [?keyword ?node ?child]
   (conda [(v+ ?node) (conda [(v+ ?keyword) (equals ?child ((?keyword (reifiers ?node))))]
                             [(v- ?keyword) (fresh [?keywords]
                                                  (equals ?keywords (keys (reifiers ?node)))
                                                  (contains ?keywords ?keyword)
                                                  (has ?keyword ?node ?child))])]
          [(v- ?node) (conda [(v+ ?child) (all (equals ?node (owner ?child)) 
                                               (has ?keyword ?node ?child))]
                             [(v- ?child) (fresh [?astkeyw]
                                              (ast ?astkeyw ?node)
                                              (has ?keyword ?node ?child))])]))

;todo: variant van conde/matche schrijven die ([:+ :-] .... neemt)


(defn child [?keyword ?node ?child]
  (fresh [?ch] 
    (has ?keyword ?node ?ch)
    (conda
           [(succeeds (instance? ASTNode ?ch)) (== ?child ?ch)] 
           [(succeeds (instance? java.util.AbstractList ?ch)) (contains ?ch ?child)])))

(defn child+ [?node ?child]
  (fresh [?keyw ?ch]
    (child ?keyw ?node ?ch)
    (conde [(== ?ch ?child)]
           [(child+ ?ch ?child)])))

;tabled version is much slower (169294ms vs 5990ms on jHotDraw)
;but might be faster if multiple child+ conditions are used in a query

;(def tabled-child+ 
;  (tabled [?node ?child]
;   (fresh [?keyw ?ch]
;          (child ?keyw ?node ?ch)
;          (conde [(== ?ch ?child)]
;                [(tabled-child+ ?ch ?child)]))))    
  

; Node generators called by reification goals
; -------------------------------------------

;TODO: keep caches in sync with workspace .. 
;perhaps add compilation unit as argument, then granularity of updates can be cus (as in cava)

;(defn lstatements []
;  (run* [?s] 
;                 (fresh [?m]
;                   (ast :MethodDeclaration ?m)
;                   (child+ ?m ?s)
;                   (succeeds (statement? ?s)))))

;(defn lexpressions []
;  (run* [?e] 
;                 (fresh [?t]
;                   (ast :TypeDeclaration ?t)
;                   (child+ ?t ?e)
;                   (succeeds (actual-expression? ?e)))))


;(def memoized-lnodes-of-type (memoize lnodes-of-type))


(defn subclass-nodes-of-type [t]
  (let [c (class-for-ekeko-keyword t)
        s (supers c)]
    (cond (contains? s org.eclipse.jdt.core.dom.Expression)
            (filter (fn [n] (instance? c n))
              (nodes-of-type :Expression))
          (contains? s org.eclipse.jdt.core.dom.Statement)
            (filter (fn [n] (instance? c n))
              (nodes-of-type :Statement))
          :else 
            (run* [?n] 
              (fresh [?cu]
                (ast :CompilationUnit ?cu)
                (child+ ?cu ?n)
                (succeeds (instance? c ?n)))))))


      
;(defmacro mapcat-key-over-mlis [key]
;  `(mapcat ~key (mlis-for-jdt)))
    

;(defn nodes-of-type [t]
;  (case t
;    :CompilationUnit (mapcat-key-over-mlis :compilationUnits)
;    :MethodDeclaration (mapcat-key-over-mlis :methodDeclarations)
;    :TypeDeclaration (mapcat-key-over-mlis :typeDeclarations)
;    :AnnnotationTypeDeclaration (mapcat-key-over-mlis :annotationTypeDeclarations)
;    :EnumDeclaration (mapcat-key-over-mlis :enumDeclarations)
;    :AnonymousClassDeclaration (mapcat-key-over-mlis :anonoymousClassDeclarations)
;    :Statement (mapcat-key-over-mlis :statements)
;    :Expression (mapcat-key-over-mlis :expressions)
;    (try (subclass-nodes-of-type t) (catch ClassNotFoundException e []))))


(defn logic-all-import-declarations []
  (run* [?importDeclaration]
        (fresh [?cu]
               (ast :CompilationUnit ?cu)
               (child :imports ?cu ?importDeclaration))))

(defn logic-all-package-declarations []
  (run* [?packageDeclaration]
        (fresh [?cu]
               (ast :CompilationUnit ?cu)
               (child :package ?cu ?packageDeclaration))))



(defn nodes-of-type [t]
  (let [models (java-project-models)]
    (case t
      :CompilationUnit (mapcat (fn [java-project-model] (.getCompilationUnits java-project-model)) models)
      :MethodDeclaration (mapcat (fn [java-project-model]  (.getMethodDeclarations java-project-model)) models)
      :TypeDeclaration (mapcat (fn [java-project-model]  (.getTypeDeclarations java-project-model)) models)
      :AnnotationTypeDeclaration (mapcat (fn [java-project-model]  (.getAnnotationTypeDeclarations java-project-model)) models)
      :EnumDeclaration (mapcat (fn [java-project-model]  (.getEnumDeclarations java-project-model)) models)
      :AnonymousClassDeclaration (mapcat (fn [java-project-model]  (.getAnonymousClassDeclarations java-project-model)) models)
      :Statement (mapcat (fn [java-project-model]  (.getStatements java-project-model)) models)
      :Expression (mapcat (fn [java-project-model]  (.getExpressions java-project-model)) models)
      :ImportDeclaration (logic-all-import-declarations)
      :PackageDeclaration (logic-all-import-declarations)
      :FieldDeclaration (mapcat (fn [java-project-model]  (.getFieldDeclarations java-project-model)) models)
      :SingleVariableDeclaration (mapcat (fn [java-project-model]  (.getSingleVariableDeclarations java-project-model)) models)
      :EnumConstantDeclaration (mapcat (fn [java-project-model]  (.getEnumConstantDeclarations java-project-model)) models)
      :AnnotationTypeMemberDeclaration (mapcat (fn [java-project-model]  (.getAnnotationTypeMemberDeclarations java-project-model)) models)
      (try (subclass-nodes-of-type t) (catch ClassNotFoundException e [])))))
                                

      
    
    
  





